==> prism-1.0.0-20240516.023415-10.jar: prism/minio.clj <==
(ns prism.minio
  (:require
    [clojure.java.io :as io]
    [hato.middleware :as hmw]
    [prism.core :refer [defdelayed] :as prism]
    [prism.internal.classpath :as cp]
    [prism.services :as services])
  (:import
    (com.github.davidmoten.aws.lw.client BaseUrlFactory Client HttpMethod Multipart Request)
    (java.net ConnectException)
    (org.json XML)))

(def ^:private dns-url-factory
  (reify
    BaseUrlFactory
    (create [_ _ _]
      (services/service-url! :minio))))

(defdelayed ^:private ^Client s3
  (let [{{:keys [username password]} :minio} (prism/config)]
    (-> (Client/s3)
        (.region "us-east-1")
        (.accessKey username)
        (.secretKey password)
        (.baseUrlFactory dns-url-factory)
        .build)))

(defn- parse-response-stream [input-stream]
  (->> input-stream io/reader XML/toJSONObject .toMap))

(defn- with-connection-handling [f]
  (try
    (f)
    (catch ConnectException _
      (services/invalidate-service! :minio)
      (when-not (services/service-url! :minio)
        (throw (ex-info "Could not resolve domain name for minio" {:status 503}))))))

(defn- make-request! [^Request request]
  (if-let [response (with-connection-handling #(.responseInputStream request))]
    (parse-response-stream response)
    (recur request)))

(defn list-objects [{:keys [bucket delimiter prefix]}]
  (-> (.path (s3) (into-array String [bucket]))
      (.query "list-type" "2")
      (.query "delimiter" delimiter)
      (.query "prefix" prefix)
      make-request!
      (get "ListBucketResult")))

(defn delete-object! [{:keys [bucket key]}]
  (-> (.path (s3) (into-array String [bucket key]))
      (.method HttpMethod/DELETE)
      make-request!))

(defn put-object! [{:keys [bucket key body]}]
  (-> (.path (s3) (into-array String [bucket key]))
      (.requestBody ^String body)
      (.method HttpMethod/PUT)
      make-request!))

(defn upload-object! [{:keys [bucket key body]}]
  (let [req (-> (Multipart/s3 (s3))
                (.bucket bucket)
                (.key key))]
    (with-connection-handling
      #(.upload req ^Callable (constantly body)))))

(defn- stream-response [^Request req]
  (let [stream (.responseInputStream req)]
    (if (-> (.statusCode stream)
            hmw/unexceptional-status?)
      stream
      (cp/when-ns 'taoensso.timbre
        (let [error (-> (parse-response-stream stream)
                        (get "Error"))]
          (when-not (->> (get error "Code")
                         (= "NoSuchKey"))
            (taoensso.timbre/warnf "Received errored minio response: '%s'" error)))))))

(defn get-object [{:keys [bucket key]}]
  (let [req (.path (s3) (into-array String [bucket key]))]
    (with-connection-handling
      #(stream-response req))))

(defn get-object-metadata [{:keys [bucket key]}]
  (with-connection-handling
    #(-> (.path (s3) (into-array String [bucket key]))
         (.method HttpMethod/HEAD)
         .response
         .headers)))

(comment
  (get-object {:bucket "ffwd-temp-test"
               :key    "some-key"})
  (put-object! {:bucket "ffwd-temp-test"
                :key    "some-key"
                :body   (slurp (io/file "deps.edn"))})
  (list-objects {:prefix    "rules/"
                 :bucket    "ffwd-loki"
                 :delimiter "/"}))

==> prism-1.0.0-20240517.034516-19.jar: prism/minio.clj <==
(ns prism.minio
  (:require
    [clojure.java.io :as io]
    [hato.middleware :as hmw]
    [prism.core :refer [defdelayed] :as prism]
    [prism.internal.classpath :as cp]
    [prism.services :as services])
  (:import
    (com.github.davidmoten.aws.lw.client BaseUrlFactory Client HttpMethod Multipart Request)
    (com.github.davidmoten.aws.lw.client.xml.builder Xml)
    (java.net ConnectException)
    (java.nio.charset StandardCharsets)
    (java.security MessageDigest)
    (java.util Base64 Map)
    (org.json XML)))

(def ^:private dns-url-factory
  (reify
    BaseUrlFactory
    (create [_ _ _]
      (services/service-url! :minio))))

(defdelayed ^:private ^Client s3
  (let [{{:keys [username password]} :minio} (prism/config)]
    (-> (Client/s3)
        (.region "us-east-1")
        (.accessKey username)
        (.secretKey password)
        (.baseUrlFactory dns-url-factory)
        .build)))

(defn- parse-response-stream [input-stream]
  (->> input-stream io/reader XML/toJSONObject .toMap))

(defn- with-connection-handling [f]
  (try
    (f)
    (catch ConnectException _
      (services/invalidate-service! :minio)
      (when-not (services/service-url! :minio)
        (throw (ex-info "Could not resolve domain name for minio" {:status 503}))))))

(defn- make-request! [^Request request]
  (if-let [response (with-connection-handling #(.responseInputStream request))]
    (parse-response-stream response)
    (recur request)))

(defn list-objects [{:keys [bucket delimiter prefix]}]
  (-> (.path (s3) (into-array String [bucket]))
      (.query "list-type" "2")
      (.query "delimiter" delimiter)
      (.query "prefix" prefix)
      make-request!
      (get "ListBucketResult")))

(defn delete-object! [{:keys [bucket key]}]
  (-> (.path (s3) (into-array String [bucket key]))
      (.method HttpMethod/DELETE)
      make-request!))

(defn put-object! [{:keys [bucket key body]}]
  (-> (.path (s3) (into-array String [bucket key]))
      (.requestBody ^String body)
      (.method HttpMethod/PUT)
      make-request!))

(defn upload-object! [{:keys [bucket key body]}]
  (let [req (-> (Multipart/s3 (s3))
                (.bucket bucket)
                (.key key))]
    (with-connection-handling
      #(.upload req ^Callable (constantly body)))))

(defn- stream-response [^Request req]
  (let [stream (.responseInputStream req)]
    (if (-> (.statusCode stream)
            hmw/unexceptional-status?)
      stream
      (cp/when-ns 'taoensso.timbre
        (let [error (-> (parse-response-stream stream)
                        (get "Error"))]
          (when-not (->> (get error "Code")
                         (= "NoSuchKey"))
            (taoensso.timbre/warnf "Received errored minio response: '%s'" error)))))))

(defn get-object [{:keys [bucket key]}]
  (let [req (.path (s3) (into-array String [bucket key]))]
    (with-connection-handling
      #(stream-response req))))

(defn get-object-metadata [{:keys [bucket key]}]
  (with-connection-handling
    #(-> (.path (s3) (into-array String [bucket key]))
         (.method HttpMethod/HEAD)
         .response
         .headers)))

(defn- delete-objects-request ^String [ks]
  (-> (reduce
        (fn [^Xml xml k]
          (-> (.element xml "Object")
              (.element "Key")
              (.content k)
              .up
              .up))
        (-> (Xml/create "Delete")
            (.attribute "xmlns" "http://s3.amazonaws.com/doc/2006-03-01/"))
        ks)
      str))

(defn- aws-md5-header [body-bytes]
  (let [md (MessageDigest/getInstance "MD5")]
    (.update md ^bytes body-bytes)
    (-> (.encode (Base64/getEncoder) (.digest md))
        (String. StandardCharsets/US_ASCII))))

(defn delete-objects! [{:keys [bucket prefix]}]
  (when-let [keys (as-> (list-objects prefix) $
                        (get $ "Contents")
                        (if (instance? Map $) (vector $) $)
                        (mapv #(get % "Key") $)
                        (seq $))]
    (let [body (-> (delete-objects-request keys)
                   .getBytes)]
      (-> (.path (s3) (into-array String [bucket "?delete"]))
          (.requestBody ^bytes body)
          (.header "Content-MD5" (aws-md5-header body))
          (.method HttpMethod/POST)
          make-request!))))

(comment
  (get-object {:bucket "ffwd-temp-test"
               :key    "some-key"})
  (put-object! {:bucket "ffwd-temp-test"
                :key    "some-key"
                :body   (slurp (io/file "deps.edn"))})
  (list-objects {:prefix    "rules/"
                 :bucket    "ffwd-loki"
                 :delimiter "/"}))
